from datetime import datetime, timedelta
import time


# Some utility classes / functions first
from threading import Thread
from typing import Iterable


class AllMatch(set):
    """Universal set - match everything"""

    def __contains__(self, item): return True


all_Match = AllMatch()


def conv_to_set(obj):  # Allow single integer to be provided
    if isinstance(obj, (int, float)):
        return set([obj])  # Single item
    if not isinstance(obj, set):
        obj = set(obj)
    return obj


# The actual Event class
class Event(object):
    def __init__(self, action,
                 minute=None,
                 hour=None,
                 day=None,
                 month=None,
                 dow=None,
                 args=(), kwargs={}):
        self.action = action
        self.args = args
        self.kwargs = kwargs

        self.minute = conv_to_set(minute or all_Match)
        self.hours = conv_to_set(hour or all_Match)
        self.days = conv_to_set(day or all_Match)
        self.months = conv_to_set(month or all_Match)
        self.dow = conv_to_set(dow or all_Match)

    def match_time(self, t):
        """Return True if this event should trigger at the specified datetime"""
        return ((t.minute in self.minute) and
                (t.hour in self.hours) and
                (t.day in self.days) and
                (t.month in self.months) and
                (t.weekday() in self.dow))

    def check(self, t):
        if self.match_time(t):
            self.action(*self.args, **self.kwargs)


class CronTab(object):
    def __init__(self, *events):
        self.events = events

    def run(self):
        t = datetime(*datetime.now().timetuple()[:5])
        while 1:
            for e in self.events:
                e.check(t)

            t += timedelta(minutes=1)
            while datetime.now() < t:
                time.sleep((t - datetime.now()).seconds)
